# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# datreant.data

"""
datreant.data --- numpy and pandas data storage for Treants in HDF5
===================================================================

"""
from datreant.core import Treant, Tree, Bundle, View, Group

from .core import DataFile
from . import pydata, npdata, pddata
from . import tests
from . import limbs
from . import agglimbs

__all__ = ['Treant', 'Group', 'Tree', 'Leaf', 'Bundle']
__version__ = "0.6.0"  # NOTE: keep in sync with RELEASE in setup.py




import datreant.core as dtr
import datreant.data.attach
import pandas as pd
import numpy as np
import pytest
import os
import py

from datreant.data.tests import test_data


class TestTreant:
    treantname = 'testtreant'
    treanttype = 'Treant'
    treantclass = dtr.treants.Treant

    @pytest.fixture
    def treant(self, tmpdir):
        with tmpdir.as_cwd():
            c = dtr.treants.Treant(TestTreant.treantname)
        return c

    class TestData:
        """Test data storage and retrieval"""

        class DataMixin:
            """Mixin class for data storage tests.

            Contains general tests to be used for all storable data formats.

            """
            handle = 'testdata'

            def test_add_data(self, treant, datastruct):
                treant.data.add(self.handle, datastruct)
                assert os.path.exists(os.path.join(treant.abspath,
                                                   self.handle,
                                                   self.datafile))

            def test_remove_data(self, treant, datastruct):
                treant.data.add(self.handle, datastruct)
                assert os.path.exists(os.path.join(treant.abspath,
                                                   self.handle,
                                                   self.datafile))

                treant.data.remove('testdata')
                assert not os.path.exists(os.path.join(treant.abspath,
                                                       self.handle,
                                                       self.datafile))

                # check that directory got deleted, too
                assert not os.path.exists(os.path.join(treant.abspath,
                                                       self.handle))

            def test_retrieve_data(self, treant, datastruct):
                treant.data.add(self.handle, datastruct)
                np.testing.assert_equal(treant.data.retrieve(self.handle),
                                        datastruct)
                np.testing.assert_equal(treant.data[self.handle],
                                        datastruct)

        class PandasMixin(DataMixin):
            """Mixin class for pandas tests"""
            datafile = datreant.data.pddata.pddatafile

            def test_retrieve_data(self, treant, datastruct):
                treant.data.add(self.handle, datastruct)
                np.testing.assert_equal(
                        treant.data.retrieve(self.handle).values,
                        datastruct.values)
                np.testing.assert_equal(
                        treant.data[self.handle].values,
                        datastruct.values)

        # TODO
        class AppendablesMixin:
            """Mixin class for pandas objects that we expect should append"""

            def test_append_data(self, treant, datastruct):
                index = datastruct.index
                for i in range(5):
                    treant.data.append(self.handle, datastruct)

                stored = treant.data.retrieve(self.handle)
                equiv = pd.concat([datastruct]*5)

                np.testing.assert_equal(stored, equiv)

        class Test_Series(test_data.Series, PandasMixin):
            pass

        class Test_DataFrame(test_data.DataFrame, PandasMixin):
            pass

        class Test_Blank_DataFrame(test_data.Blank_DataFrame, PandasMixin):
            pass

        class Test_Wide_Blank_DataFrame(test_data.Wide_Blank_DataFrame,
                                        PandasMixin):
            pass

        class Test_Thin_Blank_DataFrame(test_data.Thin_Blank_DataFrame,
                                        PandasMixin):
            pass

        class Test_Panel(test_data.Panel, PandasMixin):
            pass

        class Test_Panel4D(test_data.Panel4D, PandasMixin):
            pass

        class NumpyMixin(DataMixin):
            """Test numpy datastructure storage and retrieval"""
            datafile = datreant.data.npdata.npdatafile

        class Test_NumpyScalar(test_data.NumpyScalar, NumpyMixin):
            pass

        class Test_Numpy1D(test_data.Numpy1D, NumpyMixin):
            pass

        class Test_Numpy2D(test_data.Numpy2D, NumpyMixin):
            pass

        class Test_Wide_Numpy2D(test_data.Wide_Numpy2D, NumpyMixin):
            pass

        class Test_Thin_Numpy2D(test_data.Thin_Numpy2D, NumpyMixin):
            pass

        class Test_Numpy3D(test_data.Numpy3D, NumpyMixin):
            pass

        class Test_Numpy4D(test_data.Numpy4D, NumpyMixin):
            pass

        class PythonMixin(DataMixin):
            """Test pandas datastructure storage and retrieval"""
            datafile = datreant.data.pydata.pydatafile

            def test_overwrite_data(self, treant, datastruct):
                treant.data[self.handle] = datastruct

                # overwrite the data with a scalar
                treant.data[self.handle] = 23
                assert treant.data[self.handle] == 23

        class Test_List(test_data.List, PythonMixin):
            pass

        class Test_Dict(test_data.Dict, PythonMixin):
            pass

        class Test_Tuple(test_data.Tuple, PythonMixin):
            pass

        class Test_Set(test_data.Set, PythonMixin):
            pass

        class Test_Dict_Mix(test_data.Dict_Mix, PythonMixin):
            pass



"""
AggLimbs for convenient Treant data storage and retrieval.

"""
import six
import pandas as pd

from datreant.core.agglimbs import AggTreeLimb


class AggData(AggTreeLimb):
    """Manipulators for collection data.

    """
    _name = 'data'

    def __repr__(self):
        return "<AggData({})>".format(self.keys(scope='all'))

    def __getitem__(self, handle):
        """Retrieve aggreggated dataset from all members.

        Returns datasets indexed according to member uuids.
        See :meth:`retrieve` for more information.

        Raises :exc:`KeyError` if dataset doesn't exist for any members.

        :Arguments:
            *handle*
                name of data to retrieve; may also be a list of names

        :Returns:
            *data*
                aggregated data, indexed by member name; if *handle* was a
                list, will be a list of equal length with the aggregated
                datasets as members

        """
        if isinstance(handle, list):
            out = list()
            for item in handle:
                out.append(self.retrieve(item, by='path'))
        elif isinstance(handle, six.string_types):
            out = self.retrieve(handle, by='path')

        return out

    def keys(self, scope='all'):
        """List available datasets.

        Parameters
        ----------
        scope : {'all', 'any'}
            Keys to list. 'all' returns only handles that are present in all
            members. 'any' returns a list of all handles present in at least
            one member.

        Returns
        -------
        handles : list
            list of handles to available datasets

        """
        datasets = [set(member.data) for member in self._collection
                    if hasattr(member, 'data')]
        if datasets:
            if scope == 'any':
                out = set.union(*datasets)
            elif scope == 'all':
                out = set.intersection(*datasets)

            out = list(out)
            out.sort()
        else:
            out = list()

        return out

    @property
    def any(self):
        return self.keys('any')

    @property
    def all(self):
        return self.keys('all')

    def retrieve(self, handle, by='path', **kwargs):
        """Retrieve aggregated dataset from all members.

        This is a convenience method. The stored data structure for each member
        is read from disk and aggregated. The aggregation scheme is dependent
        on the form of the data structures pulled from each member:

        pandas DataFrames or Series
          the structures are appended together, with a new level added
          to the index giving the member (see *by*) each set of rows
          came from

        pandas Panel or Panel4D, numpy arrays, pickled python objects
          the structures are returned as a dictionary, with keys giving
          the member (see *by*) and each value giving the corresponding
          data structure

        This method tries to do smart things with the data it reads from each
        member. In particular:

        - members for which there is no data with the given handle are
          skipped

        - the lowest-common-denominator data structure is output; this
          means that if all data structures read are pandas DataFrames,
          then a multi-index DataFrame is returned; if some structures are
          pandas DataFrames, while some are anything else, a dictionary is
          returned

        :Arguments:
            *handle*
                name of data to retrieve

        :Keywords:
            *by*
                top-level index of output data structure; 'path' uses member
                path, 'name' uses member names, 'uuid' uses member uuids; if
                names are not unique, it is better to go with 'path' or 'uuid'
                ['path']

        See :meth:`datreant.data.limbs.Data.retrieve` for more information on
        keyword usage.

        :Keywords for pandas data structures:
            *where*
                conditions for what rows/columns to return
            *start*
                row number to start selection
            *stop*
                row number to stop selection
            *columns*
                list of columns to return; all columns returned by default
            *iterator*
                if True, return an iterator [``False``]
            *chunksize*
                number of rows to include in iteration; implies
                ``iterator=True``

        :Returns:
            *data*
                aggregated data structure

        """
        def dict2multiindex(agg):
            agg_mi = None
            for member in agg:
                d = agg[member]
                label = len(d.index)*[member]
                index = pd.MultiIndex.from_arrays([label, d.index])
                d.index = index

                if agg_mi is not None:
                    agg_mi = agg_mi.append(d)
                else:
                    agg_mi = d

            return agg_mi

        # first, check for existence in any member
        if handle not in self.keys('any'):
            raise KeyError(
                    "No dataset '{}' found in any member".format(handle))

        # get indexer from *by* keyword
        if by == 'path':
            def get_index(member): return member.abspath
        elif by == 'uuid':
            def get_index(member): return member.uuid
        elif by == 'name':
            def get_index(member): return member.name
            names = [member.name for member in self._collection]
            if len(set(names)) != len(names):
                raise KeyError(
                        "Member names not unique; data structure may not"
                        " look as expected. Set `by` to 'path' or 'uuid'"
                        " to avoid this.")
        else:
            raise ValueError(
                    "*by* keyword must be either 'path', 'name', or 'uuid'")

        # first, collect all the data into a dictionary, the
        # lowest-common-denominator aggregation structure
        agg = dict()
        for member in self._collection:
            try:
                agg[get_index(member)] = member.data.retrieve(handle, **kwargs)
            except (KeyError, AttributeError):
                continue

        # if data are all Series or all DataFrames, we build a multi-index
        # Series or DataFrame (respectively)
        all_s = all([isinstance(d, pd.Series) for d in agg.values()])
        all_df = all([isinstance(d, pd.DataFrame) for d in agg.values()])

        if all_s or all_df:
            agg = dict2multiindex(agg)

        return agg



"""Data structures for storage and retrieval.

"""
import datreant.core as dtr
import pandas as pd
import numpy as np
import pytest
import os
import shutil
import py


class Series():
    @pytest.fixture
    def datastruct(self):
        data = np.random.rand(10000)
        return pd.Series(data)


class DataFrame():
    @pytest.fixture
    def datastruct(self):
        data = np.random.rand(10000, 3)
        return pd.DataFrame(data, columns=('A', 'B', 'C'))


class Blank_DataFrame():
    @pytest.fixture
    def datastruct(self):
        return pd.DataFrame(np.zeros((10, 10)))


class Wide_Blank_DataFrame():
    @pytest.fixture
    def datastruct(self):
        return pd.DataFrame(np.zeros((1, 10)))


class Thin_Blank_DataFrame():
    @pytest.fixture
    def datastruct(self):
        return pd.DataFrame(np.zeros((10, 1)))


class Panel():
    @pytest.fixture
    def datastruct(self):
        data = np.random.rand(4, 10000, 3)
        return pd.Panel(data, items=('I', 'II', 'III', 'IV'),
                        minor_axis=('A', 'B', 'C'))


class Panel4D():
    @pytest.fixture
    def datastruct(self):
        data = np.random.rand(2, 4, 10000, 3)
        return pd.Panel4D(data, labels=('gallahad', 'lancelot'),
                          items=('I', 'II', 'III', 'IV'),
                          minor_axis=('A', 'B', 'C'))


class NumpyScalar():
    @pytest.fixture
    def datastruct(self):
        return np.array(20)


class Numpy1D():
    @pytest.fixture
    def datastruct(self):
        return np.random.rand(10000)


class Numpy2D():
    @pytest.fixture
    def datastruct(self):
        return np.random.rand(10000, 500)


class Wide_Numpy2D():
    @pytest.fixture
    def datastruct(self):
        return np.zeros((1, 10))


class Thin_Numpy2D():
    @pytest.fixture
    def datastruct(self):
        return np.zeros((10, 1))


class Numpy3D():
    @pytest.fixture
    def datastruct(self):
        return np.random.rand(4, 10000, 45)


class Numpy4D():
    @pytest.fixture
    def datastruct(self):
        return np.random.rand(2, 4, 10000, 45)


class List():
    @pytest.fixture
    def datastruct(self):
        return ['galahad', 'lancelot', 42, 'arthur', 3.14159]


class Dict():
    @pytest.fixture
    def datastruct(self):
        return {'pure': 'galahad', 'brave': 'lancelot', 'answer': 42,
                'king': 'arthur', 'pi-ish': 3.14159}


class Tuple():
    @pytest.fixture
    def datastruct(self):
        return ('arthur', 3.14159)


class Set():
    @pytest.fixture
    def datastruct(self):
        return {'arthur', 3.14159, 'seahorses'}


class Dict_Mix():
    @pytest.fixture
    def datastruct(self):
        return {'an array': np.random.rand(100, 46),
                'another': np.random.rand(3, 45, 2)}



"""
File backends for storing numpy arrays.

"""

import h5py

from datreant.core.backends.core import File


npdatafile = 'npData.h5'


class npDataFile(File):
    """Interface to numpy object data files.

    Data is stored as numpy arrays in the HDF5 format. This class gives the
    needed components for storing and retrieving stored data. It uses h5py as
    its backend.

    """
    def _open_file_r(self):
        return h5py.File(self.filename, 'r')

    def _open_file_w(self):
        return h5py.File(self.filename, 'w')

    def add_data(self, key, data):
        """Add a numpy array to the data file.

        If data already exists for the given key, then it is overwritten.

        :Arguments:
            *key*
                name given to the data; used as the index for retrieving
                the data later
            *data*
                the numpy array to store
        """
        with self.write():
            try:
                self.handle.create_dataset(key, data=data)
            except RuntimeError:
                del self.handle[key]
                self.handle.create_dataset(key, data=data)

    def get_data(self, key, **kwargs):
        """Retrieve numpy array stored in file.

        :Arguments:
            *key*
                name of data to retrieve

        :Returns:
            *data*
                the selected data
        """
        with self.read():
            return self.handle[key].value

    def del_data(self, key, **kwargs):
        """Delete a stored data object.

        :Arguments:
            *key*
                name of data to delete

        """
        with self.write():
            del self.handle[key]

    def list_data(self):
        """List names of all stored datasets.

        Although the true names start with '\' indicating the root of the
        HDF5 data tree, we wish to abstract this away. We remove the leading
        '\' from the output. This shouldn't cause any problems since the
        leading '\' can be neglected when referring to stored objects by name
        using all of h5py.File's methods anyway.

        """
        with self.read():
            keys = self.handle.keys()
            return keys



import random
import multiprocessing as mp
import pytest
import numpy as np
import pandas as pd

import datreant.core as dtr
import datreant.data.attach

from datreant.core.backends.statefiles import TreantFile


def append(treantfilepath, df):
    treant = dtr.Treant(treantfilepath)
    treant.data.append('testdata', df)


class TestTreantFile:

    @pytest.fixture
    def treant(self, tmpdir):
        with tmpdir.as_cwd():
            t = dtr.treants.Treant('sprout')
        return t

    @pytest.fixture
    def treantfile(self, tmpdir):
        with tmpdir.as_cwd():
            tf = TreantFile('testtreantfile.json')
        return tf

    @pytest.fixture
    def dataframe(self):
        data = np.random.rand(100, 3)
        return pd.DataFrame(data, columns=('A', 'B', 'C'))

    def test_async_append(self, treant, dataframe):
        pool = mp.Pool(processes=4)
        num = 53
        for i in range(num):
            pool.apply_async(append, args=(treant.filepath,
                                           dataframe))
        pool.close()
        pool.join()

        assert len(treant.data['testdata']) == len(dataframe)*(num+0)



import pandas as pd
import numpy as np
import pytest

import datreant.core as dtr
import datreant.data.attach
from . import test_data


class CollectionTests():
    """Tests for common elements of Group.members and Bundle"""

    @pytest.fixture
    def testtreant(self, tmpdir, request):
        with tmpdir.as_cwd():
            t = dtr.Treant('dummytreant')
        return t

    @pytest.fixture
    def testgroup(self, tmpdir, request):
        with tmpdir.as_cwd():
            g = dtr.Group('dummygroup')
            g.members.add(dtr.Treant('bark'), dtr.Treant('leaf'))
        return g

    class TestAggData:
        """Test member data functionality"""

        @pytest.fixture
        def collection(self, collection, tmpdir):
            with tmpdir.as_cwd():
                s1 = dtr.Treant('lark')
                s2 = dtr.Treant('hark')
                g3 = dtr.Group('linus')

            collection.add(s1, [g3, s2])
            return collection

        class DataMixin:
            """Mixin class for data storage tests.

            Contains general tests to be used for all storable data formats.

            """
            handle = 'testdata'

            def test_retrieve_data(self, collection, datastruct):
                agg = dict()
                for member in collection:
                    member.data.add(self.handle, datastruct)
                    agg[member.abspath] = datastruct

                np.testing.assert_equal(collection.data.retrieve(self.handle),
                                        agg)
                np.testing.assert_equal(collection.data[self.handle],
                                        agg)

        class PanelMixin:
            """Mixin class for pandas structures that don't support
            MultiIndexing.

            """
            handle = 'testdata'

            def test_retrieve_data(self, collection, datastruct):
                agg = dict()
                for member in collection:
                    member.data.add(self.handle, datastruct)
                    agg[member.abspath] = datastruct

                stored = collection.data.retrieve(self.handle)
                for item in agg:
                    np.testing.assert_equal(stored[item].values,
                                            agg[item].values)

                stored = collection.data[self.handle]
                for item in agg:
                    np.testing.assert_equal(stored[item].values,
                                            agg[item].values)

        class MultiIndexMixin(DataMixin):
            """Mixin class for pandas structures that support MultiIndexes"""

            def test_retrieve_data(self, collection, datastruct):
                agg = dict()
                for member in collection:
                    member.data.add(self.handle, datastruct)
                    agg[member.abspath] = datastruct

                def dict2multiindex(agg):
                    agg_mi = None
                    for member in agg:
                        d = agg[member].copy(deep=True)
                        label = len(d.index)*[member]
                        index = pd.MultiIndex.from_arrays([label, d.index])
                        d.index = index

                        if agg_mi is not None:
                            agg_mi = agg_mi.append(d)
                        else:
                            agg_mi = d

                    return agg_mi

                np.testing.assert_equal(
                        collection.data.retrieve(self.handle).values,
                        dict2multiindex(agg).values)
                np.testing.assert_equal(
                        collection.data[self.handle].values,
                        dict2multiindex(agg).values)

        class Test_Series(test_data.Series, MultiIndexMixin):
            pass

        class Test_DataFrame(test_data.DataFrame, MultiIndexMixin):
            pass

        class Test_Blank_DataFrame(test_data.Blank_DataFrame, MultiIndexMixin):
            pass

        class Test_Wide_Blank_DataFrame(test_data.Wide_Blank_DataFrame,
                                        MultiIndexMixin):
            pass

        class Test_Thin_Blank_DataFrame(test_data.Thin_Blank_DataFrame,
                                        MultiIndexMixin):
            pass

        class Test_Panel(test_data.Panel, PanelMixin):
            pass

        class Test_Panel4D(test_data.Panel4D, PanelMixin):
            pass

        class Test_NumpyScalar(test_data.NumpyScalar, DataMixin):
            pass

        class Test_Numpy1D(test_data.Numpy1D, DataMixin):
            pass

        class Test_Numpy2D(test_data.Numpy2D, DataMixin):
            pass

        class Test_Wide_Numpy2D(test_data.Wide_Numpy2D, DataMixin):
            pass

        class Test_Thin_Numpy2D(test_data.Thin_Numpy2D, DataMixin):
            pass

        class Test_Numpy3D(test_data.Numpy3D, DataMixin):
            pass

        class Test_Numpy4D(test_data.Numpy4D, DataMixin):
            pass

        class Test_List(test_data.List, DataMixin):
            pass

        class Test_Dict(test_data.Dict, DataMixin):
            pass

        class Test_Tuple(test_data.Tuple, DataMixin):
            pass

        class Test_Set(test_data.Set, DataMixin):
            pass

        class Test_Dict_Mix(test_data.Dict_Mix, DataMixin):
            pass


class TestBundle(CollectionTests):
    """Test Bundle features"""

    @pytest.fixture
    def collection(self):
        return dtr.Bundle()



"""
Modifications made to :mod:`datreant` classes on import of module.

"""

from datreant.core import Treant, Tree, Bundle, View, Group
from . import limbs
from . import agglimbs

Tree._attach_limb_class(limbs.Data)
Bundle._attach_agglimb_class(agglimbs.AggData)
View._attach_aggtreelimb_class(agglimbs.AggData)



"""
File backends for storing general python objects.

"""

import pickle

from datreant.core.backends.core import File


pydatafile = 'pyData.pkl'


class pyDataFile(File):
    """Interface to python object data files.

    Arbitrary python objects are stored as pickled objects on disk. This class
    gives the needed components for storing and retrieving stored data in the
    same basic way as for pandas and numpy objects. It uses pickle files for
    serialization.

    """
    def _open_file_r(self):
        return open(self.filename, 'rb')

    def _open_file_w(self):
        return open(self.filename, 'wb+')

    def add_data(self, key, data):
        """Add a numpy array to the data file.

        If data already exists for the given key, then it is overwritten.

        :Arguments:
            *key*
                not used, but needed to give consistent interface
            *data*
                the numpy array to store
        """
        with self.write():
            pickle.dump(data, self.handle)

    def get_data(self, key, **kwargs):
        """Retrieve numpy array stored in file.

        :Arguments:
            *key*
                not used, but needed to give consistent interface

        :Returns:
            *data*
                the selected data
        """
        with self.read():
            return pickle.load(self.handle)



"""
File backends for storing pandas objects.

"""

import pandas as pd
import numpy as np


from datreant.core.backends.core import File


pddatafile = 'pdData.h5'


class pdDataFile(File):
    """Interface to pandas object data files.

    Data is stored as pandas data structures (Series, DataFrame, Panel) in
    the HDF5 format. This class gives the needed components for storing
    and retrieving stored data. It uses pandas' HDFStore object as its
    backend.

    """
    def _open_file_r(self):
        return pd.HDFStore(self.filename, 'r')

    def _open_file_w(self):
        return pd.HDFStore(self.filename, 'a')

    def add_data(self, key, data):
        """Add a pandas data object (Series, DataFrame, Panel) to the data file.

        If data already exists for the given key, then it is overwritten.

        :Arguments:
            *key*
                name given to the data; used as the index for retrieving
                the data later
            *data*
                the data object to store; should be either a Series, DataFrame,
                or Panel
        """
        with self.write():
            # index all columns if possible
            try:
                # FIXME: band-aid heuristic to catch a known corner case that
                # HDFStore doesn't catch; see ``Issue 20``
                if (isinstance(data, pd.DataFrame) and
                        data.columns.dtype == np.dtype('int64')):
                    raise AttributeError

                self.handle.put(
                    key, data, format='table', data_columns=True, complevel=5,
                    complib='blosc')
            except AttributeError:
                self.handle.put(
                    key, data, format='table', complevel=5, complib='blosc')

    def append_data(self, key, data):
        """Append rows to an existing pandas data object stored in the data file.

        Note that column names of new data must match those of the existing
        data. Columns cannot be appended due to the technical details of the
        HDF5 standard. To add new columns, store as a new dataset.

        :Arguments:
            *key*
                name of existing data object to append to
            *data*
                the data object whose rows are to be appended to the existing
                stored data; must have same columns (with names) as existing
                data

        """
        with self.write():
            try:
                self.handle.append(
                    key, data, data_columns=True, complevel=5, complib='blosc')
            except AttributeError:
                self.handle.append(key, data, complevel=5, complib='blosc')

    def get_data(self, key, **kwargs):
        """Retrieve pandas object stored in file, optionally based on where criteria.

        :Arguments:
            *key*
                name of data to retrieve

        :Keywords:
            *where*
                conditions for what rows/columns to return
            *start*
                row number to start selection
            *stop*
                row number to stop selection
            *columns*
                list of columns to return; all columns returned by default
            *iterator*
                if True, return an iterator [``False``]
            *chunksize*
                number of rows to include in iteration; implies
                ``iterator=True``

        :Returns:
            *data*
                the selected data
        """
        with self.read():
            return self.handle.select(key, **kwargs)

    def del_data(self, key, **kwargs):
        """Delete a stored data object.

        :Arguments:
            *key*
                name of data to delete

        :Keywords:
            *where*
                conditions for what rows/columns to remove
            *start*
                row number to start selection
            *stop*
                row number to stop selection

        """
        with self.write():
            self.handle.remove(key, **kwargs)

    # TODO: remove this; since we only place one datastructure in an HDF5 file,
    # we don't need it
    def list_data(self):
        """List names of all stored datasets.

        Although the true names start with '\' indicating the root of the
        HDF5 data tree, we wish to abstract this away. We remove the leading
        '\' from the output. This shouldn't cause any problems since the
        leading '\' can be neglected when referring to stored objects by name
        using all of pandas.HDFStore's methods anyway.

        """
        with self.read():
            keys = self.handle.keys()
            return [i.lstrip('/') for i in keys]



"""
Limbs for convenient Treant data storage and retrieval.

"""
import os
import six
from functools import wraps

from datreant.core.limbs import TreeLimb
from . import npdata, pddata, pydata
from .core import DataFile


class Data(TreeLimb):
    """Interface to stored data.

    """
    _name = 'data'

    def __repr__(self):
        return "<Data({})>".format(self.keys())

    def __str__(self):
        data = self.keys()
        agg = "Data"
        majsep = "="
        seplength = len(agg)

        if not data:
            out = "No Data"
        else:
            out = agg + '\n'
            out = out + majsep * seplength + '\n'
            for datum in data:
                out = out + "'{}'\n".format(datum)
        return out

    def __iter__(self):
        return self.keys().__iter__()

    def _makedirs(self, p):
        """Make directories and all parents necessary.

        :Arguments:
            *p*
                directory path to make
        """
        try:
            os.makedirs(p)
        except OSError:
            pass

    def _get_datafile(self, handle):
        """Return path to datafile corresponding to given handle.

        :Arguments:
            *handle*
                name of dataset whose datafile path to return

        :Returns:
            *datafile*
                datafile path; None if does not exist
            *proxyfile*
                proxyfile path; None if does not exist
            *datafiletype*
                datafile type; either ``persistence.pddatafile`` or
                ``persistence.npdatafile``

        """
        datafile = None
        datafiletype = None
        for dfiletype in (pddata.pddatafile, npdata.npdatafile,
                          pydata.pydatafile):
            dfile = os.path.join(self._tree.abspath,
                                 handle, dfiletype)
            if os.path.exists(dfile):
                datafile = dfile
                proxyfile = os.path.join(self._tree.abspath,
                                         handle, ".{}.proxy".format(dfiletype))
                datafiletype = dfiletype

        if datafile is None and datafiletype is None:
            raise KeyError("No data for '{}'".format(handle))

        return (datafile, proxyfile, datafiletype)

    def _read_datafile(func):
        """Decorator for generating DataFile instance for reading data.

        DataFile instance is generated and mounted at self._datafile. It
        is then dereferenced after the method call. Since data files can be
        deleted in the filesystem, this should handle cleanly the scenarios
        in which data appears, goes missing, etc. while a Treant is loaded.

        .. note:: Methods wrapped with this decorator need to have *handle* as
                  the first argument.

        """
        @wraps(func)
        def inner(self, handle, *args, **kwargs):
            filename, proxy, filetype = self._get_datafile(handle)

            if filename:
                self._datafile = DataFile(
                        os.path.join(self._tree.abspath,
                                     handle),
                        datafiletype=filetype)
                try:
                    out = func(self, handle, *args, **kwargs)
                finally:
                    del self._datafile
            else:
                out = None

            return out

        return inner

    def _write_datafile(func):
        """Decorator for generating DataFile instance for writing data.

        DataFile instance is generated and mounted at self._datafile. It
        is then dereferenced after the method call. Since data files can be
        deleted in the filesystem, this should handle cleanly the scenarios
        in which data appears, goes missing, etc. while a Treant is loaded.

        .. note:: Methods wrapped with this decorator need to have *handle* as
                  the first argument.

        """
        @wraps(func)
        def inner(self, handle, *args, **kwargs):
            dirname = os.path.join(self._tree.abspath, handle)

            self._makedirs(dirname)
            self._datafile = DataFile(dirname)

            try:
                out = func(self, handle, *args, **kwargs)
            finally:
                del self._datafile

            return out

        return inner

    def __getitem__(self, handle):
        """Get dataset corresponding to given handle(s).

        If dataset doesn't exist, ``None`` is returned.

        :Arguments:
            *handle*
                name of data to retrieve; may also be a list of names

        :Returns:
            *data*
                stored data; if *handle* was a list, will be a list
                of equal length with the stored data as members; will yield
                ``None`` if requested data is nonexistent

        """
        if isinstance(handle, list):
            out = list()
            for item in handle:
                out.append(self.retrieve(item))
        elif isinstance(handle, six.string_types):
            out = self.retrieve(handle)

        return out

    def __setitem__(self, handle, data):
        """Set dataset corresponding to given handle.

        A data instance must be either a pandas Series, DataFrame, or Panel
        object. If dataset doesn't exist, it is added. If a dataset already
        exists for the given handle, it is replaced.

        :Arguments:
            *handle*
                name given to data; needed for retrieval
            *data*
                data to store; must be a pandas Series, DataFrame, or Panel

        """
        self.add(handle, data)

    def __delitem__(self, handle):
        """Remove a dataset.

        Note: the directory containing the dataset file (``Data.h5``) will NOT
        be removed if it still contains file after the removal of the dataset
        file.

        :Arguments:
            *handle*
                name of dataset to delete

        """
        self.remove(handle)

    @_write_datafile
    def add(self, handle, data):
        """Store data in Treant.

        A data instance can be a pandas object (Series, DataFrame, Panel),
        a numpy array, or a pickleable python object. If the dataset doesn't
        exist, it is added. If a dataset already exists for the given handle,
        it is replaced.

        :Arguments:
            *handle*
                name given to data; needed for retrieval
            *data*
                data structure to store

        """
        self._datafile.add_data('main', data)

    def remove(self, handle, **kwargs):
        """Remove a dataset, or some subset of a dataset.

        Note: in the case the whole dataset is removed, the directory
        containing the dataset file (``Data.h5``) will NOT be removed if it
        still contains file(s) after the removal of the dataset file.

        For pandas objects (Series, DataFrame, or Panel) subsets of the whole
        dataset can be removed using keywords such as *start* and *stop* for
        ranges of rows, and *columns* for selected columns.

        :Arguments:
            *handle*
                name of dataset to delete

        :Keywords:
            *where*
                conditions for what rows/columns to remove
            *start*
                row number to start selection
            *stop*
                row number to stop selection
            *columns*
                columns to remove

        """
        datafile, proxy, datafiletype = self._get_datafile(handle)

        if kwargs and datafiletype == pddata.pddatafile:
            self._delete_data(handle, **kwargs)
        elif datafile:
            os.remove(datafile)
            os.remove(proxy)
            top = self._tree.abspath
            directory = os.path.dirname(datafile)
            while directory != top:
                try:
                    os.rmdir(directory)
                    directory = os.path.dirname(directory)
                except OSError:
                    break

    @_write_datafile
    def _delete_data(self, handle, **kwargs):
        """Remove a dataset, or some subset of a dataset.

        This method loads the given data instance before doing anything,
        and should generally not be used to remove data since it will create
        a datafile object if one is not already present, which could have
        side-effects for other instances of this Treant.

        Note: in the case the whole dataset is removed, the directory
        containing the dataset file (``Data.h5``) will NOT be removed if it
        still contains file(s) after the removal of the dataset file.

        :Arguments:
            *handle*
                name of dataset to delete

        :Keywords:
            *where*
                conditions for what rows/columns to remove
            *start*
                row number to start selection
            *stop*
                row number to stop selection
            *columns*
                columns to remove

        """
        # only called for pandas objects at the moment
        filename, proxy, filetype = self._get_datafile(handle)
        self._datafile.datafiletype = filetype
        try:
            self._datafile.del_data('main', **kwargs)
        except NotImplementedError:
            datafile = self._get_datafile(handle)[0]

            os.remove(datafile)
            os.remove(proxy)
            top = self._tree.abspath
            directory = os.path.dirname(datafile)
            while directory != top:
                try:
                    os.rmdir(directory)
                    directory = os.path.dirname(directory)
                except OSError:
                    break

    @_read_datafile
    def retrieve(self, handle, **kwargs):
        """Retrieve stored data.

        The stored data structure is read from disk and returned.

        If dataset doesn't exist, ``None`` is returned.

        For pandas objects (Series, DataFrame, or Panel) subsets of the whole
        dataset can be returned using keywords such as *start* and *stop* for
        ranges of rows, and *columns* for selected columns.

        Also for pandas objects, the *where* keyword takes a string as input
        and can be used to filter out rows and columns without loading the full
        object into memory. For example, given a DataFrame with handle 'mydata'
        with columns (A, B, C, D), one could return all rows for columns A and
        C for which column D is greater than .3 with::

            retrieve('mydata', where='columns=[A,C] & D > .3')

        Or, if we wanted all rows with index = 3 (there could be more than
        one)::

            retrieve('mydata', where='index = 3')

        See :meth:`pandas.HDFStore.select` for more information.

        :Arguments:
            *handle*
                name of data to retrieve

        :Keywords:
            *where*
                conditions for what rows/columns to return
            *start*
                row number to start selection
            *stop*
                row number to stop selection
            *columns*
                list of columns to return; all columns returned by default
            *iterator*
                if True, return an iterator [``False``]
            *chunksize*
                number of rows to include in iteration; implies
                ``iterator=True``

        :Returns:
            *data*
                stored data; ``None`` if nonexistent

        """
        return self._datafile.get_data('main', **kwargs)

    @_write_datafile
    def append(self, handle, data):
        """Append rows to an existing dataset.

        The object must be of the same pandas class (Series, DataFrame, Panel)
        as the existing dataset, and it must have exactly the same columns
        (names included).

        :Arguments:
            *handle*
                name of data to append to
            *data*
                data to append

        """
        self._datafile.append_data('main', data)

    def keys(self):
        """List available datasets.

        :Returns:
            *handles*
                list of handles to available datasets

        """
        datasets = list()
        top = self._tree.abspath
        for root, dirs, files in os.walk(top):
            if ((pddata.pddatafile in files) or
                    (npdata.npdatafile in files) or
                    (pydata.pydatafile in files)):
                datasets.append(os.path.relpath(root, start=top))

        datasets.sort()

        return datasets



"""
Abstract interface components for reading and writing datasets.

"""

import os

import numpy as np
import pandas as pd

from . import pydata
from . import npdata
from . import pddata


class DataFile(object):
    """Interface to data files.

    This is an abstraction layer to the pddata.pdDataFile, npdata.npDataFile,
    and pydata.pyDataFile objects. This can be used by higher level objects
    without worrying about whether to use pandas storers, numpy storers, or
    pickle.

    """

    def __init__(self, datadir, datafiletype=None, **kwargs):
        """Initialize data interface.

        :Arguments:
           *datadir*
              path to data directory
           *datafiletype*
              If known, either pddata.pddatafile or npdata.npdatafile

        """
        self.datadir = datadir
        self.datafile = None

        # if given, can get data
        self.datafiletype = datafiletype

    def add_data(self, key, data):
        """Add a pandas data object (Series, DataFrame, Panel), numpy array,
        or pickleable python object to the data file.

        If data already exists for the given key, then it is overwritten.

        :Arguments:
            *key*
                name given to the data; used as the index for retrieving
                the data later
            *data*
                the data object to store; should be either a pandas Series,
                DataFrame, Panel, or a numpy array
        """
        if isinstance(data, np.ndarray):
            self.datafile = npdata.npDataFile(
                os.path.join(self.datadir, npdata.npdatafile))
        elif isinstance(data, (pd.Series, pd.DataFrame, pd.Panel, pd.Panel4D)):
            self.datafile = pddata.pdDataFile(
                os.path.join(self.datadir, pddata.pddatafile))
        else:
            self.datafile = pydata.pyDataFile(
                os.path.join(self.datadir, pydata.pydatafile))

        self.datafile.add_data(key, data)

        # dereference
        self.datafile = None

    def append_data(self, key, data):
        """Append rows to an existing pandas data object stored in the data file.

        Note that column names of new data must match those of the existing
        data. Columns cannot be appended due to the technical details of the
        HDF5 standard. To add new columns, store as a new dataset.

        :Arguments:
            *key*
                name of existing data object to append to
            *data*
                the data object whose rows are to be appended to the existing
                stored data; must have same columns (with names) as existing
                data

        """
        # TODO: add exceptions where appending isn't possible
        if isinstance(data, np.ndarray):
            raise TypeError('Cannot append numpy arrays.')
        elif isinstance(data, (pd.Series, pd.DataFrame, pd.Panel, pd.Panel4D)):
            self.datafile = pddata.pdDataFile(
                os.path.join(self.datadir, pddata.pddatafile))
            self.datafile.append_data(key, data)
        else:
            raise TypeError('Cannot append python object.')

            # dereference
            self.datafile = None

    def get_data(self, key, **kwargs):
        """Retrieve data object stored in file.

        :Arguments:
            *key*
                name of data to retrieve

        :Keywords:
            *where*
                for pandas objects, conditions for what rows/columns to return
            *start*
                for pandas objects, row number to start selection
            *stop*
                for pandas objects, row number to stop selection
            *columns*
                for pandas objects, list of columns to return; all columns
                returned by default
            *iterator*
                for pandas objects, if True, return an iterator [``False``]
            *chunksize*
                for pandas objects, number of rows to include in iteration;
                implies ``iterator=True``

        :Returns:
            *data*
                the selected data
        """
        if self.datafiletype == npdata.npdatafile:
            self.datafile = npdata.npDataFile(
                os.path.join(self.datadir, npdata.npdatafile))
            out = self.datafile.get_data(key, **kwargs)
            self.datafile = None
        elif self.datafiletype == pddata.pddatafile:
            self.datafile = pddata.pdDataFile(
                os.path.join(self.datadir, pddata.pddatafile))
            out = self.datafile.get_data(key, **kwargs)
            self.datafile = None
        elif self.datafiletype == pydata.pydatafile:
            self.datafile = pydata.pyDataFile(
                os.path.join(self.datadir, pydata.pydatafile))
            out = self.datafile.get_data(key)
            self.datafile = None
        else:
            raise TypeError('Cannot return data without knowing datatype.')
            out = None

        return out

    def del_data(self, key, **kwargs):
        """Delete a stored data object.

        :Arguments:
            *key*
                name of data to delete

        :Keywords:
            *where*
                for pandas objects, conditions for what rows/columns to remove
            *start*
                for pandas objecs, row number to start selection
            *stop*
                for pandas objects, row number to stop selection

        """
        if self.datafiletype == npdata.npdatafile:
            self.datafile = npdata.npDataFile(
                os.path.join(self.datadir, npdata.npdatafile))
            out = self.datafile.del_data(key, **kwargs)
            self.datafile = None
        elif self.datafiletype == pddata.pddatafile:
            self.datafile = pddata.pdDataFile(
                os.path.join(self.datadir, pddata.pddatafile))
            out = self.datafile.del_data(key, **kwargs)
            self.datafile = None
        elif self.datafiletype == pydata.pydatafile:
            pass
        else:
            raise TypeError('Cannot return data without knowing datatype.')
            out = None

        return out



#! /usr/bin/python
"""Setuptools-based setup script for datreant.data.

For a basic installation just type the command::

  python setup.py install

"""

from setuptools import setup, find_packages

setup(name='datreant.data',
      version='0.6.0',
      description='convenient data storage and retrieval in HDF5 for Treants',
      author='David Dotson',
      author_email='dotsdl@gmail.com',
      url = 'http://datreant.org/',
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],
      packages=find_packages('src'),
      namespace_packages=['datreant'],
      package_dir={'': 'src'},
      scripts=[],
      license='BSD',
      long_description=open('README.rst').read(),
      install_requires=[
          'datreant.core',
          'six',
          'numpy',
          'pandas',
          'tables',
          'h5py',
          ]
      )
